from cml.domain.deconstruction.abstract import Abstract
from cml.shared.errors import (
    NotEnoughFeaturesWarning,
    ReconstructionFailure,
    WinnerSelectionFailure)

__all__ = (
    "ZSigmaDeconstructor",
)


class ZSigmaDeconstructor(Abstract):
    RELATION = ("Sigma", "Z")
    MIN_N_FEATURE = 2

    def __init__(self, *args):
        super().__init__(*args)

    def deconstruct(self, relative, log):
        if self.time_constraint(relative):
            overlapping = self.identify_overlapping_features(relative)
            if self.column_constraint(overlapping):
                meta_model = self.model_fusion(relative)
                return self.reconstructing(overlapping, meta_model, relative, log)

        return log

    def time_constraint(self, relative):
        if self.settings.deconst_max_distance_t == 0:
            return self.overlapping_time(self.pragmatic, relative)

        elif 0 < self.settings.deconst_max_distance_t < 1:
            return self.enclosing_time(self.pragmatic, relative)

        elif self.settings.deconst_max_distance_t == 1:
            return True

    def overlapping_time(self, prag_model, relative_model):
        return relative_model.min_timestamp >= prag_model.max_timestamp or \
            prag_model.min_timestamp >= relative_model.max_timestamp

    def enclosing_time(self, prag_model, relative_model):
        m_dash_max_time = max(prag_model.max_timestamp,
                              relative_model.max_timestamp)
        m_dash_min_time = min(prag_model.min_timestamp,
                              relative_model.min_timestamp)
        relative_m_dash_time = self.settings.deconst_max_distance_t*(
                m_dash_max_time-m_dash_min_time)

        return (relative_model.min_timestamp - prag_model.max_timestamp) < \
               relative_m_dash_time or (prag_model.min_timestamp -
                                        relative_model.max_timestamp) < relative_m_dash_time

    def identify_overlapping_features(self, relative):
        relative_block = relative.trained_with(self.knowledge)
        overlapped_block = self.block.same_features_fusion(relative_block)
        return overlapped_block

    def column_constraint(self, block):
        return block.n_columns(effective=True) >= self.MIN_N_FEATURE

    def reconstructing(self, block, meta_model, relative, log):
        try:
            ml_models = meta_model.subject
            _, pragmatics = self.reconstructor.reconstruct(
                self.level, block, ml_models, meta_model)
            winner = pragmatics.pop()[0]
            # winner, _ = determine_winner({reliability: pragmatics})

            winner.identifier = self.correct_reconstructed_identifier(self.level)
            self.knowledge.replace(relative, winner)
            assert winner.level == self.level

            log.reconstructed_info.append(winner)
            log.inserted.append(winner)
            log.deleted.append(relative)
            log.status = True
            return log

        except (ReconstructionFailure,
                NotEnoughFeaturesWarning,
                WinnerSelectionFailure):
            relative_block = relative.trained_with(self.knowledge)
            modification = self.model_disposal(self.pragmatic, relative,
                                               block, relative_block)
            log.inserted.append(modification.inserted)
            log.deleted.append(modification.deleted)
            return log

